#!/bin/sh /usr/share/centrifydc/perl/run

################################################################################
#
# Copyright (C) 2014-2020 Centrify Corporation. All rights reserved.
#
# This script helps to clean up Centrify DirectControl Agent core files older
# than n days.
#
# Generally, the core files are located under process's current working
# directory, and named as core*. See the following for details:
# Linux: 
#    default: process's current working directory.
#    command: sysctl -w kernel.core_pattern="<core dir>" 
#    config: /proc/sys/kernel/core_pattern
#    note: program uses the core location configuration before it was crashed.
#          The core file might be redirected to:
#          |/usr/libexec/abrt-hook-ccpp %s %c %p %u %g %t e
#    (Test on RHEL3 RHEL6, Ubuntu12, Fedora18, SUSE9)
#
# AIX:
#    default: process's current working directory.
#    command: chcore
#    config:   /etc/security/user:
#              root:
#              core_pathname = /cores
#              core_path = on
#    note: program use the core location configuration when it was started.
#          core file will always be started with "core", we cannot change it.
#    (AIX5.3)
#
# SOLARIS:
#   default: process's current working directory.
#   command: coreadm
#   config: /etc/coreadm.conf: if *_GLOB_* not enabled, use the *_INIT_*.
#           COREADM_GLOB_PATTERN=/var/tmp/core.%p
#           COREADM_GLOB_ENABLED=yes
#           COREADM_INIT_PATTERN=core
#   note: program uses the core location configuration before it was crashed.
#    (Sol10)
#         Solaris 11.1 doesn't have file /etc/coreadm.conf.
#
# Mac OS X:
#    default:  /cores/
#    not configurable.
#
# HPUX:
#    Default: process's current working directory. The core file will be named
#             as "core". And the new core file will cover the old one.
#    not configurable.
# 
#
################################################################################
use strict;
use lib '/usr/share/centrifydc/perl';
use CentrifyDC::Include;
use CentrifyDC::Platform qw(GetPlatform OS_FAMILY);
use CentrifyDC::GP::Args;
use CentrifyDC::GP::RegHelper;
use CentrifyDC::GP::General qw(:debug ReadFile RunCmd IsEmpty EscapeSingleQuote);
use File::Basename;


# >>> DATA >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

my $REGKEY = "software/policies/centrify/centrifydc/settings/adclient";
my $REGVALUE = "adclient.coredump.cleanup";
my $CDC_LOCATION = "/var/centrifydc";
my @CDC_DAEMONS = (
    'adclient',
    'cdcwatch',
    'kcm',
);
my $CORE_PATTERN = "core";
my $CORE_FILE_PERMISSION = 0600;
my $CMD_FILE = "/usr/bin/file";
# Hash to record the adclient core file on darwin. Key is pid of crashed adclient
my %darwin_core;

sub GetCoreDirAndPattern($);
sub GetLinuxDumpCoreLocation();
sub GetCorePathAndSwitchFromStanza($);
sub GetAixRootDumpCoreConfiguration($);
sub GetAixDumpCoreLocation();
sub InitDarwinCrashCoreReports();
sub GetSolarisDumpCoreLocation();
sub GetCoreDumpInfo($);
sub IsDarwinCentrifyCoreDump($);
sub IsCentrifyCoreDump($);
sub ClearCoreDumps($);

# >>> SUB >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# Get the dir name from a string
# If the string is not a valid path, return the adclient working directory:
# /etc/centrifydc
#
# $_[0]: $path - path string.
#
# return $dir, $pattern - $dir: the directory of dump core location.
#                         $pattern: the core file pattern
#                         undef: failure
#
sub GetCoreDirAndPattern($)
{
    my $path = shift;
    # Remove the trailing or starting spaces.
    # If the pattern is spaces, script will not work.
    $path =~ s/^\s//g;
    $path =~ s/\s$//g;
    # If meet a relative path, then convert it to absolute path.
    if ($path !~ '^/')
    {
        $path = "$CDC_LOCATION/$path";
    }
    
    my ($pattern, $dir, $suffix) = fileparse($path);
    if (-d $dir)
    {
        # Remove the ending '/' if have.
        $dir =~ s/\/$//;
        return ($dir, $pattern);
    }
    return undef;
}

# Get the configured dump core location on Linux platform.
# Generally, it's configured by 'sysctl -w var=value' and pushed into
# "/proc/sys/kernel/core_pattern"
# e.g. /tmp/core.%p.%g
#      /tmp/core.%p.%g|/usr/libexec/abrt-hook-ccpp %s %c %p %u %g %t e
#
# return $dir, $pattern - $dir: the directory of dump core location.
#                         $pattern: the core file pattern
#
sub GetLinuxDumpCoreLocation()
{
    # This file should be maintained by sysctl. Assume that there's no comment
    # and empty lines
    my $core_conf = "/proc/sys/kernel/core_pattern";
    my $config = ReadFile($core_conf);
    chomp $config;
    TRACE_OUT("Get the core dump configuration: [$config]");
    # Remove the redirection string.
    # e.g. |/usr/libexec/abrt-hook-ccpp %s %c %p %u %g %t e
    $config =~ s/\|.*$//;
    return GetCoreDirAndPattern($config);
}

#
# Get the dump core switch and core path from a /etc/security/user stanza
#
#   $_[0]: $fd - the file descriptor.
#
#   return ($switch, $path) - the dump core switch and core path.
#       $1: on|off|undef
#       $2: <core path>|undef
#
sub GetCorePathAndSwitchFromStanza($)
{
    my $FD = shift;
    # core_path = "on"
    my $switch_pattern = qr/^\s*core_path\s*=[\s"']*([^\s'"]*)/;
    # core_pathname = "/cores.%p"
    my $corepath_pattern = qr/^\s*core_pathname\s*=[\s"']*([^\s'"]*)/;
    my ($switch, $path);
    my $cur_pos = tell($FD);
    while (<$FD>)
    {
        # Skip the comment.
        next if (/^\*/);
        # Blank line. End of stanza. 
        last if (/^\s*$/);
        # Meet the next stanza, push the line back to <$FD>.
        # Actually, if the file is healthy, we will not step into this
        # statement.
        if (/[^#:]*:/)
        {
            seek($FD, $cur_pos, 0);
            last;
        }
        if (/$switch_pattern/)
        {
            # The first setting win.
            $switch = $1 if (!$switch);
        }
        if (/$corepath_pattern/)
        {
            # The first setting win.
            $path = $1 if (!$path);
        }
        $cur_pos = tell($FD);
    }

    return ($switch, $path);
}

# Get the dump core directory or user 'root' from configuration file.
# If core_path is on, then try to fetch from the core_pathname.
# Otherwise return the adclient working directory: /etc/centrifydc
# root:
#    core_pathname = /cores.%p
#    core_path = on
#
# $_[0]: $path - the absolute path of the configuration file.
#
# return: $path - the path of root dump core directory.
#
sub GetAixRootDumpCoreConfiguration($)
{
    my $path = shift;
    my $find_root = 0;
    my $config = "";
    my $rc = open my $FD, '<', $path;
    if (!$rc)
    {
        DEBUG_OUT("Cannot open [$path] to read: $!");
        DEBUG_OUT("Use default dump core location: [$CDC_LOCATION]");
        return "$CDC_LOCATION";
    }
    my $root_core_switch;
    my $root_core_path;
    my $default_core_switch;
    my $default_core_path;
    while (<$FD>)
    {
        my $line = $_;
        # Skip the comment.
        next if ($line =~ qr/^\*/);

        # Note: The valid stanza head is a user name, followed by a :, not allowed
        # empty string before user name or colon symbol.
        # i.e. "   root:", "root :"

        # The stanza of user root
        if ($line =~ qr/^root:/)
        {
            # The first setting wins.
            my ($switch, $corepath) = GetCorePathAndSwitchFromStanza($FD);
            $root_core_switch = $switch if (!$root_core_switch);
            $root_core_path = $corepath if (!$root_core_path);
        }

        # The stanza of default setting.
        if ($line =~ qr/^default:/)
        {
            # The first setting wins.
            my ($switch, $corepath) = GetCorePathAndSwitchFromStanza($FD);
            $default_core_switch = $switch if (!$default_core_switch);
            $default_core_path = $corepath if (!$default_core_path);
        }
    }
    close $FD;

    # When stanzas conflict occurred, 'root' wins.
    TRACE_OUT("Dump core path enabled of root: [$root_core_switch], default: [$default_core_switch]");
    if ($root_core_switch eq 'on' or $default_core_switch eq 'on')
    {
        TRACE_OUT("Dump core path of root: [$root_core_path], default: [$default_core_path]");
        return $root_core_path if ($root_core_path);
        return $default_core_path if ($default_core_path);
    }
    else
    {
        # If core_paths are off, use the current working directory.
        return "$CDC_LOCATION";
    }
}

# Get the configured dump core location on Aix platform.
# Generally, it's configured by command 'chcore' and pushed into
# "/etc/security/user".
#
# return $dir, $pattern - $dir: the directory of dump core location.
#                         $pattern: the core file pattern
#
sub GetAixDumpCoreLocation()
{
    my $core_conf = "/etc/security/user";
    my $pattern = GetAixRootDumpCoreConfiguration($core_conf);
    TRACE_OUT("Get the core dump configuration: [$pattern]");
    # Always be a path.
    $pattern .= '/';
    my ($dir, $core_pattern) = GetCoreDirAndPattern($pattern);
    # Aix core file name will always be started with "core"
    return ($dir, $CORE_PATTERN);
}

# On Mac OS X, every coredump will have a *.crash report file under /Library/Logs/DiagnosticReports/
# Check all the .crash there, and store the core file pid of adclient/cdcwatch to hash.
# e.g. core file [/cores/core.75069] will have:
#      /Library/Logs/DiagnosticReports/adclient_2014-09-01-140331_flueg-mb1083.crash
#      line 1: Process:         adclient [75069]
#
#   return:     1 - success
#           undef - fail
#
sub InitDarwinCrashCoreReports()
{
    my $report_dir = "/Library/Logs/DiagnosticReports";
    my $rc = opendir(my $FD, $report_dir);
    if (! $rc)
    {
        ERROR_OUT("Cannot open directory [$report_dir] to read: $!");
        return undef;
    }
    # Fetch the core file records' path
    my $daemons_pattern = join '|', @CDC_DAEMONS;
    $daemons_pattern = "($daemons_pattern)";
    # Note: the crash file name will be started with daemon's name. So we can
    # filter the files here.
    my @core_files = grep {/$daemons_pattern.*crash$/} readdir($FD);
    closedir $FD;

    foreach my $cfile (@core_files)
    {
        # Fetch the first line content.
        my $file_path = "$report_dir/$cfile";
        my $cmd = ["/usr/bin/head", "-n", "1", $file_path];
        my ($rc, $line) = RunCmd($cmd);
        foreach my $PATTERN (@CDC_DAEMONS)
        {
            # Get the pid from report.
            # We want to strictly match the word.
            if ($line =~ /.*(\b$PATTERN\b)\s*\[(\d+)\]/)
            {
                $darwin_core{$2} = 1;
                TRACE_OUT("Centrify daemon [$1] had a crash report with pid [$2]");
            }
        }
    }
    return 1;
}

# Get the configured dump core location on SunOS platform.
# Generally, it's configured by command 'coreadm' and pushed into
# "/etc/coreadm.conf".
# Since the /etc/coreadm.conf are not always available from all Solaris
# revision, get configuration from the command output of 'coreadm'.
# 1. If the global setting is enabled, then fetch the GLOB_PATTERN.
# 2. If not, then fetch the INIT_PATTERN.
# 3. If the PATTERN has a valid directory, return it. Otherwise return the
#    adclient working directory: /etc/centrifydc
# command output:
#           global core file pattern: /cores
#           global core dumps: disabled
#           init core file pattern: core
#
# return $dir, $pattern - $dir: the directory of dump core location.
#                         $pattern: the core file pattern
#
sub GetSolarisDumpCoreLocation()
{
    my $core_cmd = "/usr/bin/coreadm";
    my $pattern = "";
    my ($rc, $config) = RunCmd($core_cmd);
    if (!$rc)
    {
        # Pattern "global core dumps: disabled"
        my $glob_enabled = ($config =~ qr/.*global core dumps: ([^\s]*)\s*/)[0];
        if ($glob_enabled eq "enabled")
        {
            # Pattern "global core file pattern: /cores"
            $pattern = ($config =~ qr/.*global core file pattern: ([^\s]*)\s*/)[0];
            TRACE_OUT("coreadm global pattern enabled: [$pattern]");
        }
        else
        {
            # Pattern "init core file pattern: core"
            $pattern = ($config =~ qr/.*init core file pattern: ([^\s]*)\s*/)[0];
        }
    }
    TRACE_OUT("Get the core dump configuration: [$pattern]");
    return GetCoreDirAndPattern($pattern);
}

# Return the adclient dump core directory depending on diff platform.
#
#   $_[0]: $platform - OS platform project.
#
# return $dir, $pattern - $dir: the directory of dump core location.
#                         $pattern: the core file pattern
#
sub GetCoreDumpInfo($)
{
    my $platform = shift;
    my $os_family = $platform->getOSFamily();

    if ($os_family eq OS_FAMILY->{LINUX})
    {
        return GetLinuxDumpCoreLocation();
    }
    elsif ($os_family eq OS_FAMILY->{AIX})
    {
        return GetAixDumpCoreLocation();
    }
    elsif ($os_family eq OS_FAMILY->{DARWIN})
    {
        InitDarwinCrashCoreReports();
        # On Mac OS X, the dump core directory is fixed to '/cores'.
        $CORE_FILE_PERMISSION = 0400;
        return ("/cores", "$CORE_PATTERN");
    }
    elsif ($os_family eq OS_FAMILY->{SOLARIS})
    {
        return GetSolarisDumpCoreLocation();
    }
    elsif ($os_family eq OS_FAMILY->{HPUX})
    {
        # On HP-UX, the dump core directory is the one adclient working on.
        return ("$CDC_LOCATION", "$CORE_PATTERN");
    }
    else
    {
        # For unknown platform, nothing to do.
        return ("$CDC_LOCATION", "$CORE_PATTERN");
    }

}

# Check whether the core file is owned by adclient
#
#   $_[0]: $path - the file path of core file.
#
#   return: 1 - yes
#           0 - no
sub IsDarwinCentrifyCoreDump($)
{
    my $core_file = shift;
    my $pid = ($core_file =~ /.*core\.([\d]+)/)[0];
    if ($pid)
    {
        return 1 if ($darwin_core{$pid});
    }
    return 0;
}

#
# Check whether the core file is owned by adclient
#
#   $_[0]: $path - the file path of core file.
#
#   return:     1 - yes
#               0 - no
#           undef - error, should skip this core.
#
sub IsCentrifyCoreDump($)
{
    my $path = shift;
    # $ sudo file /var/centrifydc/core.3533
    # /var/centrifydc/core.3533: ELF 64-bit LSB core file x86-64, version 1 (SYSV),
    # SVR4-style, from '/usr/sbin/adclient'
    my $cmd = [$CMD_FILE, $path];
    my ($rc, $output) = RunCmd($cmd);
    if (!$rc)
    {
        foreach my $PATTERN (@CDC_DAEMONS)
        {
            # We want to strictly match the word.
            if ($output =~ /(\b$PATTERN\b)/)
            {
                DEBUG_OUT("Core dump [$path] was created by Centrify daemon [$1]");
                return 1;
            }
        }
        DEBUG_OUT("Core dump [$path] is not created by Centrify daemons");
        return 0;
    }
    else
    {
        # We don't want to exit if fail to parse the owner of one core.
        ERROR_OUT("Failed to check the owner of core [$path], skip it.");
        return undef;
    }
}

# Main subroutine to clear the dump cores of adclient.
#
#   $_[0]: $threshold - clean up core dumps older than $threshold days.
#
#   return:    1 - success
#          undef - fail
#
sub ClearCoreDumps($)
{
    my $cleanup_threshold = shift;
    DEBUG_OUT("Clean up cores dumped more than [$cleanup_threshold] days ago.");
    my $platform = GetPlatform();
    my $os_family = $platform->getOSFamily();

    my ($core_dir, $file_pattern) = GetCoreDumpInfo($platform);
    if (!$core_dir)
    {
        ERROR_OUT("Cannot fetch the core dump location. Stop processing.");
        return undef;
    }
    DEBUG_OUT("Core dump dir: [$core_dir] Pattern [$file_pattern]");

    # If the core dump location isn't /var/centrifydc, need to check whether the
    # core are owned by Centrify daemons.
    if ($os_family ne OS_FAMILY->{DARWIN} and
        "$core_dir" ne "$CDC_LOCATION" and
        (! -x $CMD_FILE))
    {
        # Will not process if cannot verify the core dumps' owner.
        ERROR_OUT("Cannot find command [$CMD_FILE]. So will not clean up core dumps in custom core dump directory [$core_dir]");
        return undef;
    }

    my $rc = opendir(my $FD, $core_dir);
    if (! $rc)
    {
        ERROR_OUT("Cannot open directory [$core_dir] to read: $!");
        return undef;
    }
    # Fetch the dump core file paths
    my @core_files = grep {/^$file_pattern.*/} readdir($FD);
    closedir $FD;

    foreach my $cfile (@core_files)
    {
        my $file_path = "$core_dir/$cfile";
        my ($mode, $uid, $mtime) = (stat($file_path))[2, 4, 9];
        $mode &= 07777;
        DEBUG_OUT("process file [$cfile] mode [$mode], uid [$uid], mtime [$mtime]");
        if ($mode != $CORE_FILE_PERMISSION)
        {
            DEBUG_OUT("File permission isn't correct, should be [%04o] instead of [%04o]. Skip it.", $CORE_FILE_PERMISSION, $mode);
            next;
        }
        if ($uid != 0)
        {
            DEBUG_OUT("File ownership isn't correct, should be [0] instead of [$uid]. Skip it.");
            next;
        }

        # We will clear all the core dumps under /var/centrifydc.
        # Otherwise, we only clear the core dumps by adclient/cdcwatch/centrify-kcm
        if ("$core_dir" ne "$CDC_LOCATION")
        {
            # Check if the core file owner is adclient or not.
            $rc = 0;
            if ($os_family eq OS_FAMILY->{DARWIN})
            {
                $rc = IsDarwinCentrifyCoreDump($cfile);
            }
            else
            {
                $rc = IsCentrifyCoreDump($file_path);
                # Skip if cannot parse the core file.
                next if (! defined($rc));
            }
            if (! $rc)
            {
                DEBUG_OUT("Core file isn't owned by Centrify daemons. Skip it.");
                next;
            }
        }

        my $cur_time = time();
        # Get the age (in second) of core file since it was created.
        my $core_age = $cur_time - $mtime;
        next if ($core_age <= 0);

        if ($core_age > ($cleanup_threshold * 3600 * 24))
        {
            # If the core is a link, just delete as well.
            INFO_OUT("Delete core file [$file_path] created [%d] days before", $core_age/(3600 * 24));
            unlink $file_path or ERROR_OUT("Failed to delete file [$file_path]: $!");
        }
    }
    return 1;
}

# >>> MAIN >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

my $args = CentrifyDC::GP::Args->new();
# Do nothing when 'unmap'
$args->isMap() or exit 0;
    
# Load the machine registry
CentrifyDC::GP::Registry::Load(undef);

my $reg = CentrifyDC::GP::RegHelper->new($args->action(), $args->class(), $REGKEY, undef, undef);
$reg or FATAL_OUT("Cannot create RegHelper instance");
$reg->load('current');

# Do nothing when gp disabled
my $cur_reg = $reg->get('current');
if (defined $cur_reg)
{
    my $cleanup_threshold = $cur_reg->{$REGVALUE};
    if (defined $cleanup_threshold)
    {
        if ($cleanup_threshold > 0)
        {
            ClearCoreDumps($cleanup_threshold);
        }
        else
        {
            ERROR_OUT("The cleanup threshold is invalid: [$cleanup_threshold]");
        }
    }
}
