################################################################################
#
# Copyright (C) 2013-2020 Centrify Corporation. All rights reserved.
#
#
# Centrify DirectControl platform detection module.
#
# This module provides functions to retrieve platform information such as OS
# type, OS major & minor versions, and architecture.
#
# Supported platforms include:
#     RHEL, Fedora, CentOS, Ubuntu, Debian, Solaris, AIX, HP-UX, Darwin
#
# The logic to detect platform information is from install.sh. But there is no
# need to follow install.sh to filter out old OS versions. Because whether the
# platform is supported by Centrify products is determined during installation,
# not runtime.
#
# Note: This module should be kept synchronized with
# cli\surveyor\base\lib\Centrify\Platform.pm for maintenance purpose.
#
################################################################################


package CentrifyDC::Platform;

use strict;
use base 'Exporter';
use CentrifyDC::GP::General qw(:debug RunCommand ReadFile);

our @EXPORT_OK = qw(GetPlatform OS_FAMILY OS ARCH);

our $FILE_OS_RELEASE = '/etc/os-release';
our %OS_RELEASE = ();

# >>> DATA >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>


use constant OS_FAMILY =>
{
    UNKNOWN     => 'unknown',
    LINUX       => 'linux',
    SOLARIS     => 'solaris',
    AIX         => 'aix',
    HPUX        => 'hpux',
    DARWIN      => 'darwin'
};

use constant OS =>
{
    UNKNOWN     => 'unknown',
    ORACLE      => 'oracle',
    RHEL        => 'rhel',
    FEDORA      => 'rhfc',
    CENTOS      => 'centos',
    SCIENTIFIC  => 'scientific',
    MANDRIVA    => 'mandriva',
    XEN         => 'xenserver',
    SUSE        => 'suse',
    OPENSUSE    => 'opensuse',
    UBUNTU      => 'ubuntu',
    DEBIAN      => 'debian',
    AIX         => 'aix',
    SOLARIS     => 'solaris',
    HPUX        => 'hpux',
    DARWIN      => 'darwin',
    AMAZONAMI   => 'amazonami',
    AMAZON      => 'amazon',
    COREOS      => 'coreos',
    RHELAH      => 'rhelah'
};

use constant OS_EDITION =>
{
    UNKNOWN    => 'unknown',   # unknown edition
    SERVER     => 'server',    # server edition
    DESKTOP    => 'desktop',   # desktop edition
    STANDARD   => 'standard'   # edition that not distinguish server and desktop
};

use constant ARCH =>
{
    UNKNOWN     => 'unknown',
    I386        => 'i386',
    X86_64      => 'x86_64',
    POWERPC     => 'powerpc',
    SPARC       => 'sparc',
    PA          => 'pa',
    IA64        => 'ia64',
    POWERPC64LE => 'powerpc64le'
};

my $uname;

# >>> SUB >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>


# public class subroutines
sub GetPlatform;

# private class subroutines
sub _DetectOSFamily;
sub _DetectOSRev;
sub _DetectArch;
sub _DetectOSRevLinux;
sub _DetectArchLinux;
sub _DetectOSRevAix;
sub _DetectArchAix;
sub _DetectOSRevSolaris;
sub _DetectArchSolaris;
sub _DetectOSRevHpux;
sub _DetectArchHpux;
sub _DetectOSRevDarwin;
sub _DetectArchDarwin;

# public member subroutines
sub getOSFamily;
sub getOS;
sub getOSRevFull;
sub getOSRevMajor;
sub getOSRevMinor;
sub getOSEdition;
sub getArch;

# private member subroutines
sub _new;


#
# Detect OS family. It is a dispatcher to functions to do detection.
#
#   return: string
#
sub _DetectOSFamily
{
    my $os_family = OS_FAMILY->{UNKNOWN};

    my ($rc, $output) = RunCommand("$uname -s");
    return OS_FAMILY->{UNKNOWN} if ($rc ne '0');

    if (index($output, 'Linux') != -1)
    {
        $os_family = OS_FAMILY->{LINUX};
    }
    elsif (index($output, 'SunOS') != -1)
    {
        $os_family = OS_FAMILY->{SOLARIS};
    }
    elsif (index($output, 'AIX') != -1)
    {
        $os_family = OS_FAMILY->{AIX};
    }
    elsif (index($output, 'HP-UX') != -1)
    {
        $os_family = OS_FAMILY->{HPUX};
    }
    elsif (index($output, 'Darwin') != -1)
    {
        $os_family = OS_FAMILY->{DARWIN};
    }
    else
    {
        WARN_OUT("All known unix names cannot be found");
        DEBUG_OUT("uname -s output: [$output]");
    }

    return $os_family;
}

#
# Detect OS type and revision. Major and minor numbers will be provided.
#
# Note that the detection logic in this function and osutil.cpp
# should be in synchronization with that in install.sh. Hence, 
# any changes in the detection logic made to this function, should be also 
# made to osutil.cpp.
#
# Note that the detection logic of OS edition in Platform.pm and osutil.cpp 
# is not fully consistent to install.sh.
#
#   $_[0]:  OS family
#
#   return: OS type
#           OS major rev
#           OS minor rev
#           OS edition
#
sub _DetectOSRev
{
    my $os_family = $_[0];
    my $os = OS->{UNKNOWN};
    my $os_rev_major = 0;
    my $os_rev_minor = 0;
    my $os_edition = OS_EDITION->{UNKNOWN};

    if ($os_family eq OS_FAMILY->{LINUX})
    {
        ($os, $os_rev_major, $os_rev_minor, $os_edition) = _DetectOSRevLinux();
    }
    elsif ($os_family eq OS_FAMILY->{AIX})
    {
        ($os, $os_rev_major, $os_rev_minor, $os_edition) = _DetectOSRevAix();
    }
    elsif ($os_family eq OS_FAMILY->{SOLARIS})
    {
        ($os, $os_rev_major, $os_rev_minor, $os_edition) = _DetectOSRevSolaris();
    }
    elsif ($os_family eq OS_FAMILY->{HPUX})
    {
        ($os, $os_rev_major, $os_rev_minor, $os_edition) = _DetectOSRevHpux();
    }
    elsif ($os_family eq OS_FAMILY->{DARWIN})
    {
        ($os, $os_rev_major, $os_rev_minor, $os_edition) = _DetectOSRevDarwin();
    }
    else
    {
        WARN_OUT("OS family [$os_family] not supported");
    }

    return ($os, $os_rev_major, $os_rev_minor, $os_edition);
}

#
# Detect platform architecture.
#
#   $_[0]:  OS family
#
#   return: architecture found
#
sub _DetectArch
{
    my $os_family = $_[0];
    my $arch = ARCH->{UNKNOWN};

    if ($os_family eq OS_FAMILY->{LINUX})
    {
        $arch = _DetectArchLinux();
    }
    elsif ($os_family eq OS_FAMILY->{AIX})
    {
        $arch = _DetectArchAix();
    }
    elsif ($os_family eq OS_FAMILY->{SOLARIS})
    {
        $arch = _DetectArchSolaris();
    }
    elsif ($os_family eq OS_FAMILY->{HPUX})
    {
        $arch = _DetectArchHpux();
    }
    elsif ($os_family eq OS_FAMILY->{DARWIN})
    {
        $arch = _DetectArchDarwin();
    }
    else
    {
        WARN_OUT("OS family [$os_family] not supported");
    }

    return $arch;
}

sub _DetectOSRevAix
{
    my $os = OS->{UNKNOWN};
    my $os_rev_major = 0;
    my $os_rev_minor = 0;
    my $os_edition = OS_EDITION->{UNKNOWN};
    
    my ($rc, $output) = RunCommand("$uname -rv");
    if ($rc eq '0')
    {
        chomp $output;
        my ($minor, $major) = split ' ', $output;
        if (defined $minor and defined $major)
        {
            $os = OS->{AIX};
            $os_rev_major = $major;
            $os_rev_minor = $minor;
            $os_edition = OS_EDITION->{STANDARD};  
        }
    }
    return ($os, $os_rev_major, $os_rev_minor, $os_edition);
}

sub _DetectArchAix
{
    my $arch = ARCH->{UNKNOWN};
    my ($rc, $output) = RunCommand("$uname -p");
    return ARCH->{UNKNOWN} if ($rc ne '0');
    $arch = ARCH->{POWERPC} if ($output =~ /powerpc/);

    # IBM AIX supported platforms: ROMP, IBM POWER, PowerPC, IBM PS/2,
    # System/370, ESA/390
    # But we only support the PowerPC architectures.

    return $arch;
}

#
# Read os-release file
#
# Read /etc/os-release file and store the data to global hash %OS_RELEASE
#
# See more info about the os-release file:
# https://www.freedesktop.org/software/systemd/man/os-release.html
#
sub _ReadOSReleaseFile
{
    if (-f $FILE_OS_RELEASE || -l $FILE_OS_RELEASE)
    {
        my $output = ReadFile($FILE_OS_RELEASE);
        my @lines = split(/\n/, $output);
        foreach (@lines) {
            chomp;              # no newline
            s/#.*//;            # no comments
            s/^\s+//;           # no leading white
            s/\s+$//;           # no trailing white
            next unless length; # anything left?
            my ($var, $value) = split(/\s*=\s*/, $_, 2);
            if ($var ne "" and defined $value)
            {
                if ($value ne "")
                {
                    # Remove quotes
                    if ($value =~ /^"([^"]*)"$/)
                    {
                        $value = $1;
                    }
                    elsif ($value =~ /^'([^']*)'$/)
                    {
                        $value = $1;
                    }
                }
                $OS_RELEASE{$var} = $value;
            }
        }
    }
}

sub _DetectOSRevLinux2
{
    my $os = OS->{UNKNOWN};
    my $os_rev_major = 0;
    my $os_rev_minor = 0;
    my $os_edition = OS_EDITION->{UNKNOWN};

    _ReadOSReleaseFile();

    if (defined $OS_RELEASE{NAME})
    {
        if ($OS_RELEASE{NAME} =~ /^Amazon Linux AMI/)
        {
            $os = OS->{AMAZONAMI};
            $os_edition = OS_EDITION->{STANDARD};
        }
        elsif ($OS_RELEASE{NAME} eq 'Amazon Linux')
        {
            $os = OS->{AMAZON};
            $os_edition = OS_EDITION->{STANDARD};
        }
        elsif ($OS_RELEASE{NAME} =~ /^CentOS/)
        {
            $os = OS->{CENTOS};
            $os_edition = OS_EDITION->{STANDARD};
        }
        elsif ($OS_RELEASE{NAME} =~ /^Container Linux by CoreOS/)
        {
            $os = OS->{COREOS};
            $os_edition = OS_EDITION->{STANDARD};
        }
        elsif ($OS_RELEASE{NAME} =~ /^Red Hat Enterprise Linux Atomic Host/)
        {
            $os = OS->{RHELAH};
            $os_edition = OS_EDITION->{STANDARD};
        }
        elsif ($OS_RELEASE{NAME} =~ /^Debian/)
        {
            $os = OS->{DEBIAN};
            $os_edition = OS_EDITION->{STANDARD};
        }
        elsif ($OS_RELEASE{NAME} =~ /^Fedora/)
        {
            $os = OS->{FEDORA};
            $os_edition = OS_EDITION->{STANDARD};
        }
        elsif ($OS_RELEASE{NAME} =~ /^Linux Mint LMDE/)
        {
            $os = OS->{UBUNTU};
            $os_edition = OS_EDITION->{STANDARD};
        }
        elsif ($OS_RELEASE{NAME} =~ /^openSUSE/)
        {
            $os = OS->{OPENSUSE};
            $os_edition = OS_EDITION->{STANDARD};
        }
        elsif ($OS_RELEASE{NAME} =~ /^Scientific Linux/)
        {
            $os = OS->{SCIENTIFIC};
            $os_edition = OS_EDITION->{STANDARD};
        }
        elsif ($OS_RELEASE{NAME} =~ /^Oracle Linux/)
        {
            $os = OS->{ORACLE};
            $os_edition = OS_EDITION->{STANDARD};
        }
        elsif ($OS_RELEASE{NAME} =~ /^Red Hat Enterprise Linux/)
        {
            $os = OS->{RHEL};
            $os_edition = OS_EDITION->{STANDARD};
        }
        elsif ($OS_RELEASE{NAME} =~ /^SLED/)
        {
            $os = OS->{SUSE};
            $os_edition = OS_EDITION->{DESKTOP};
        }
        elsif ($OS_RELEASE{NAME} =~ /^SLES/)
        {
            $os = OS->{SUSE};
            $os_edition = OS_EDITION->{SERVER};
        }
        elsif ($OS_RELEASE{NAME} =~ /^Ubuntu/)
        {
            $os = OS->{UBUNTU};
            $os_edition = OS_EDITION->{STANDARD};
        }
        else
        {
            DEBUG_OUT("Fail to detect OS version from %s", $FILE_OS_RELEASE);
        }

        if ($os ne OS->{UNKNOWN} and defined $OS_RELEASE{VERSION_ID})
        {

            # VERSION_ID is mostly numeric. It can be as simple as "7"
            # or contains character other than numeric, e.g. "_" or "-".
            #
            # Description from official:
            # A lower-case string (mostly numeric, no spaces or other
            # characters outside of 0-9, a-z, ".", "_" and "-") identifying
            # the operating system version, excluding any OS name information
            # or release code name, and suitable for processing by scripts or
            # usage in generated filenames. This field is optional.
            # Example: "VERSION_ID=17" or "VERSION_ID=11.04".
            #
            # Parse the VERSION_ID and get the major and minor numbers
            # from the beginning.
            #
            my @rev_numbers;
            $_ = ".$OS_RELEASE{VERSION_ID}";
            while (/\G\.(\d+)/g)
            {
                push(@rev_numbers, $1);
            }

            # Set major and minor if it is defined.
            $os_rev_major = $rev_numbers[0] if (defined $rev_numbers[0]);
            $os_rev_minor = $rev_numbers[1] if (defined $rev_numbers[1]);
        }
    }

    return ($os, $os_rev_major, $os_rev_minor, $os_edition);
}

sub _DetectOSRevLinux
{
    my ($os, $os_rev_major, $os_rev_minor, $os_edition) = _DetectOSRevLinux2();
    if ($os ne OS->{UNKNOWN})
    {
        return ($os, $os_rev_major, $os_rev_minor, $os_edition);
    }

    my $FILE_ENTERPRISE_RELEASE = '/etc/enterprise-release';
    my $FILE_ORACLE_RELEASE = '/etc/oracle-release';
    my $FILE_REDHAT_RELEASE = '/etc/redhat-release';
    my $FILE_SUSE_RELEASE = '/etc/SuSE-release';
    my $FILE_UNITEDLINUX_RELEASE = '/etc/UnitedLinux-release';
    my $FILE_LSB_RELEASE = '/etc/lsb-release';
    my $FILE_DEBIAN_VERSION = '/etc/debian_version';
    my $FILE_SYSTEM_RELEASE = '/etc/system-release';

    # oracle linux
    if (-f $FILE_ENTERPRISE_RELEASE ||
        -f $FILE_ORACLE_RELEASE &&
        -f $FILE_REDHAT_RELEASE)
    {
        my $output;
        if (-f $FILE_ENTERPRISE_RELEASE)
        {
            $output = ReadFile($FILE_ENTERPRISE_RELEASE);
        }
        else
        {
            $output = ReadFile($FILE_REDHAT_RELEASE);
        }
        
        # Oracle Linux 4.8
        # (1) content of "/etc/enterprise-release":
        #     Enterprise Linux Enterprise Linux AS release 4 (October Update 8)
        # (2) content of "/etc/redhat-release":
        #     Red Hat Enterprise Linux AS release 4 (Nahant Update 8)
        # (3) "/etc/oracle-release" doesn't exist 
        
        # Oracle Linux 5.8        
        # (1) content of "/etc/enterprise-release":
        #     Enterprise Linux Enterprise Linux Server release 5.8 (Carthage)
        # (2) content of "/etc/redhat-release":
        #     Red Hat Enterprise Linux Server release 5.8 (Tikanga)
        # (3) content of "/etc/oracle-release":
        #     Oracle Linux Server release 5.8
        
        # Oracle Linux 6.6
        # (1) "/etc/enterprise-release" doesn't exist
        # (2) content of "/etc/redhat-release":
        #     Red Hat Enterprise Linux Server release 6.6 (Santiago)
        # (3) content of "/etc/oracle-release":
        #     Oracle Linux Server release 6.6
        
        # Oracle Linux 7
        # (1) "/etc/enterprise-release" doesn't exist
        # (2) content of "/etc/redhat-release":
        #     Red Hat Enterprise Linux Server release 7.0 (Maipo)
        # (3) content of "/etc/oracle-release":
        #     Oracle Linux Server release 7.0
        
        if (index($output, 'Enterprise Linux') != -1)
        {
            if (index($output, 'release') != -1)
            {
                my ($major, $minor) = ($output =~ /release\s+(\d+)\.(\d+)/);
                
                if (!defined($major))
                {
                    ($major) = ($output =~ /release\s+(\d+)/);
                    ($minor) = ($output =~ /Update\s+(\d+)/);
                }
                
                if (defined($major))
                {
                    $os = OS->{ORACLE};
                    $os_rev_major = $major;
                    $os_rev_minor = $minor if defined($minor);
                    
                    # detect os edition
                    my ($edition) = ($output =~ /Enterprise Linux\s+(.+)\s+release/);
                    if ($os_rev_major > 4)
                    {
                        if (defined($edition))
                        {
                            if (index($edition, 'Server') != -1)
                            {
                                $os_edition = OS_EDITION->{SERVER};                        
                            }
                            elsif (index($edition, 'Client') != -1)
                            {
                                $os_edition = OS_EDITION->{DESKTOP};   
                            }
                            else
                            { 
                                # We don't consider unknown edition as unknown platform.
                                # Platform can still be a valid platform even if the edition is unknown.
                                $os_edition = OS_EDITION->{UNKNOWN};
                            }
                        }
                        else
                        {
                            $os_edition = OS_EDITION->{UNKNOWN};
                        }
                    }
                    else
                    {
                        # oracle 4 are considered as server edition
                        $os_edition = OS_EDITION->{SERVER};
                    }
                }
            }
        }
    }
    elsif ( -f $FILE_REDHAT_RELEASE) # Red Hat
    {
        my $output = ReadFile($FILE_REDHAT_RELEASE);

        if (index($output, 'Enterprise Linux') != -1)
        {
            if (index($output, 'release') != -1)
            {
                # Normal case:
                # - Red Hat Enterprise Linux Server release 5.8 (Tikanga)
                # - Red Hat Enterprise Linux Client release 6.2 (Santiago)
                
                my ($major, $minor) = ($output =~ /release\s+(\d+)\.(\d+)/);
                
                # Not always have minor version before bracket (e.g. RHEL3):
                # - Red Hat Enterprise Linux WS release 3 (Taroon Update 6)
                #
                # Not always have minor version (e.g. RHEL 4.0):
                # - Red Hat Enterprise Linux ES release 4 (Nahant)
                
                if (!defined($major))
                {
                    ($major) = ($output =~ /release\s+(\d+)/);
                    ($minor) = ($output =~ /Update\s+(\d+)/);
                }

                if (defined($major))
                {
                    $os = OS->{RHEL};
                    $os_rev_major = $major;
                    $os_rev_minor = $minor if defined($minor);
   
                    # detect os edition
                    my ($edition) = ($output =~ /Enterprise Linux\s+(.+)\s+release/);
                    if ($os_rev_major > 4)
                    {
                        if (defined($edition))
                        {
                            if (index($edition, 'Server') != -1)
                            {
                                $os_edition = OS_EDITION->{SERVER};                        
                            }
                            elsif (index($edition, 'Client') != -1)
                            {
                                $os_edition = OS_EDITION->{DESKTOP};   
                            }
                            else
                            {
                                # We don't consider unknown edition as unknown platform.
                                # Platform can still be a valid platform even if the edition is unknown.
                                $os_edition = OS_EDITION->{UNKNOWN};
                            }
                        }
                        else
                        {
                            $os_edition = OS_EDITION->{UNKNOWN}; 
                        }
                    }
                    else
                    {
                        # rhel 3 and 4 are considered as server edition
                        $os_edition = OS_EDITION->{SERVER};    
                    }
                }
            }
        }
        elsif (index($output, 'Fedora') != -1)
        {
            # Content of the release file looks like:
            #     Fedora release 8 (Werewolf)
            # or
            #     Fedora release 14 (Laughlin)
            #
            if (index($output, 'release') != -1)
            {
                my ($major) = ($output =~ /release\s+(\d+)/);
                # For Fedora, we support release 14 or later.
                if ($major > 13)
                {
                    $os = OS->{FEDORA};
                    $os_rev_major = $major;
                    $os_edition = OS_EDITION->{STANDARD};
                }
                else
                {
                    DEBUG_OUT("[Fedora release %s] detected, does not support", $major);
                }
            }
        }
        elsif (index($output, 'CentOS') != -1)
        {
            # Conent of the release file looks like: 
            #     CentOS release 6.5 (Final)
            
            if (index($output, 'release') != -1)
            {
                my ($major, $minor) = ($output =~ /release\s+(\d+)\.(\d+)/);   # \surveyor\base\lib\Centrify\Platform.pm should be corrected like this
                ($major) = ($output =~ /release\s+(\d+)/) if (!defined($major));
                if (defined($major))
                {
                    $os = OS->{CENTOS};
                    $os_rev_major = $major;
                    # CentOS might not have minor version.
                    $os_rev_minor = $minor if defined($minor);
                    $os_edition = OS_EDITION->{STANDARD};
                }
            }
        }
        elsif (index($output, 'Scientific Linux') != -1)
        {
            # Content of the release file looks like:
            #    Scientific Linux release 6.3 (Carbon)
            
            if (index($output, 'release') != -1)
            {
                my ($major, $minor) = ($output =~ /release\s+(\d+)\.(\d+)/);
                ($major) = ($output =~ /release\s+(\d+)/) if (!defined($major));
                if (defined($major))
                {
                    $os = OS->{SCIENTIFIC};
                    $os_rev_major = $major;
                    $os_rev_minor = $minor if defined($minor);
                    $os_edition = OS_EDITION->{STANDARD};
                }
            }
        }
        elsif (index($output, 'Mandriva Linux') != -1)
        {
            # Content of the release file looks like:
            #     Mandriva Linux release 2010
            # or
            #     Mandriva Linux Enterprise Server release 5           

            if (index($output, 'release') != -1)
            {
                my ($major, $minor) = ($output =~ /release\s+(\d+)\.(\d+)/);
                ($major) = ($output =~ /release\s+(\d+)/) if (!defined($major));

                if (defined($major))
                {
                    $os = OS->{MANDRIVA};
                    $os_rev_major = $major;
                    $os_rev_minor = $minor if defined($minor);
                    
                    # detect os edition
                    my ($edition) = ($output =~ /Mandriva Linux\s+(.+)\s+release/);
                    if (defined($edition))
                    {
                        if (index($edition, 'Server') != -1)
                        {
                            $os_edition = OS_EDITION->{SERVER};
                        }
                        else
                        {
                            $os_edition = OS_EDITION->{UNKNOWN};
                        }
                    }
                    else
                    {
                        $os_edition = OS_EDITION->{DESKTOP};
                    }
                }
            }
        }
        elsif (index($output, 'XenServer') != -1)
        {
            # Content of the release file looks like:
            #     XenServer release 6.2.0-70446c (xenenterprise)
            
            my ($major, $minor) = ($output =~ /release\s+(\d+)\.(\d+).+xenenterprise/);
            ($major) = ($output =~ /release\s+(\d+).+xenenterprise/) if (!defined($major));
            if (defined($major))
            {
                $os = OS->{XEN};
                $os_rev_major = $major;
                $os_rev_minor = $minor if defined($minor);
                $os_edition = OS_EDITION->{STANDARD};
            }
        }
        else
        {
            # Amazon Linux
            if (-f $FILE_SYSTEM_RELEASE)
            {
                my $output = ReadFile($FILE_SYSTEM_RELEASE);
                if (index($output, 'Amazon Linux AMI') != -1)
                {
                    # Content of the release file looks like:
                    #     Amazon Linux AMI release 2015.09

                    # This means _DetectOSRevLinux2() failed to find /etc/os-release file so this is old Amazon Linux
                    $os = OS->{AMAZONAMI};
                    $os_edition = OS_EDITION->{STANDARD};
                    # TBD: Parse the version number from the string
                }
            }
            else
            {
                DEBUG_OUT("Fail to detect OS version from %s", $FILE_REDHAT_RELEASE);
            }
        }
    }
    elsif (-f $FILE_SUSE_RELEASE)  # suse
    {
        my $output = ReadFile($FILE_SUSE_RELEASE);
        
        # Content of this release file looks like,
        #
        # For OpenSuse, it looks like
        #     openSUSE 12.1 (i586)
        #     VERSION = 12.1
        #     CODENAME = Asparagus
        # or
        #     openSUSE 12.2 (x86_64)
        #     VERSION = 12.2
        #     CODENAME = Mantis
        #  
        # For suse9 or above, it looks like
        #     SUSE LINUX Enterprise Server 9 (x86_64)
        #     VERSION = 9
        #     PATCHLEVEL|RELEASE =
        #
        # For suse8, it looks like
        #     SuSE SLES-8 (i386)
        #     VERSION = 8.1
        #
        
        if ($output =~ /open/i) #openSuse
        {
            my ($major, $minor) = ($output =~ /VERSION\s+=\s+(\d+)\.(\d+)/);
            ($major) = ($output =~ /VERSION\s+=\s+(\d+)/) if (!defined($major));
            if (defined($major))
            {
                $os = OS->{OPENSUSE};
                $os_rev_major = $major;
                $os_rev_minor = $minor if defined($minor);
                $os_edition = OS_EDITION->{STANDARD};
            }
        }
        elsif ($output =~ /suse/i) # SUSE Linux Enterprise Server/Desktop
        {
            my ($major, $minor) = ($output =~ /VERSION\s+=\s+(\d+)\.(\d+)/);
            ($major) = ($output =~ /VERSION\s+=\s+(\d+)/) if (!defined($major));
            if (defined($major))
            {
                $os = OS->{SUSE};
                $os_rev_major = $major;
                $os_rev_minor = $minor if defined($minor);
                
                # detect os edition
                if (index($output, 'Desktop') != -1)
                {
                    $os_edition = OS_EDITION->{DESKTOP};
                }
                elsif (index($output, 'Server') != -1)
                {
                    $os_edition = OS_EDITION->{SERVER};
                }
                elsif (index($output, 'SLES') != -1)
                {
                    $os_edition = OS_EDITION->{SERVER};
                }
                else
                {
                    $os_edition = OS_EDITION->{UNKNOWN};
                }
            }
        }
        else
        {
            DEBUG_OUT("Fail to detect OS version from %s", $FILE_SUSE_RELEASE);
        }
    }
    elsif (-f $FILE_UNITEDLINUX_RELEASE)  # suse8
    {
        my $output = ReadFile($FILE_UNITEDLINUX_RELEASE);
        
        # Content of this release file looks like:
        #     UnitedLinux 1.0 (i586)
        #     VERSION = 1.0
        
        if ($output =~ /VERSION = 1\./)
        {
            $os = OS->{SUSE};
            $os_rev_major = 8;
            $os_rev_minor = 0;
            $os_edition = OS_EDITION->{STANDARD};
        }
        else
        {
            DEBUG_OUT("Fail to detect OS version from %s", $FILE_UNITEDLINUX_RELEASE);
        }
    }
    elsif (-f $FILE_LSB_RELEASE) # Mint or Ubuntu
    {
        my $output = ReadFile($FILE_LSB_RELEASE);
        
        # Content of this release file looks like:
        #     DISTRIB_ID=Ubuntu
        #     DISTRIB_RELEASE=12.10
        #     DISTRIB_CODENAME=quantal
        #     DISTRIB_DESCRIPTION="Ubuntu 12.10"
        #
        
        if ($output =~ /DISTRIB_ID=Ubuntu/) # Ubuntu
        {
            my ($major, $minor) = ($output =~ /DISTRIB_RELEASE=(\d+)\.(\d+)/);
            if (defined($major) and defined($minor))
            {
                $os = OS->{UBUNTU};
                # Simplify the Ubuntu 6 to 6.04 and 6.06 release
                if ($major == 6)
                {
                    if ($minor > 6)
                    {
                        $major++;
                        $minor = 4;
                    }
                    else
                    {
                        $minor = 6;
                    }
                }
                # Simplify the Ubuntu 7 or later to xx.04 and xx.10 release
                else
                {
                    if ($minor > 10)
                    {
                        $major++;
                        $minor = 4;
                    }
                    elsif ($minor < 5)
                    {
                        $minor = 4;
                    }
                    else
                    {
                        $minor = 10;
                    }
                }
                $os_rev_major = $major;
                $os_rev_minor = $minor;
                $os_edition = OS_EDITION->{STANDARD};
            }
        }
        else
        {
            DEBUG_OUT("Fail to detect OS version from %s", $FILE_LSB_RELEASE);
        }
    }
    # Debian
    elsif (-f $FILE_DEBIAN_VERSION)
    {
        my $output = ReadFile($FILE_DEBIAN_VERSION);
        my ($major, $minor) = ($output =~ /(\d+)\.(\d+)/);
        if (defined($major) and defined($minor))
        {
            if ($major > 4)
            {
                $os = OS->{DEBIAN};
                $os_rev_major = $major;
                $os_rev_minor = $minor;
                $os_edition = OS_EDITION->{STANDARD};
            }
            else
            {
                DEBUG_OUT("Debian 4 or earlier version detected, does not support");
            }
        }
        else
        {
            DEBUG_OUT("Fail to detect OS version from %s", $FILE_DEBIAN_VERSION);
        }
    }
    else
    {
        WARN_OUT("All known version files of Linux distributions cannot be found");
    }

    return ($os, $os_rev_major, $os_rev_minor, $os_edition);
}

sub _DetectArchLinux
{
    my $arch = ARCH->{UNKNOWN};

    my ($rc, $output) = RunCommand("$uname -m");
    return ARCH->{UNKNOWN} if ($rc ne '0');

    $arch = ARCH->{I386} if ($output =~ /i686/);
    $arch = ARCH->{X86_64} if ($output =~ /x86_64/);
    if ($output =~ /ppc64le/)
    {
        $arch = ARCH->{POWERPC64LE};
    }
    elsif ($output =~ /ppc64/)
    {
        $arch = ARCH->{POWERPC};
    }
    return $arch;
}

sub _DetectOSRevSolaris
{
    my $os = OS->{UNKNOWN};
    my $os_rev_major = 0;
    my $os_rev_minor = 0;
    my $os_edition = OS_EDITION->{UNKNOWN};
    
    my ($rc, $output) = RunCommand("$uname -r");
    if ($rc eq '0')
    {
        chomp $output;
        TRACE_OUT("Output of uname -r: $output");
        # Solaris does not have a minor version before solaris11
        # e.g. 5.10 should be solaris10
        my ($legacy, $major) = split '\.', $output;
        if (defined $major)
        {
            $os = OS->{SOLARIS};
            $os_rev_major = $major;
            $os_edition = OS_EDITION->{STANDARD};
            # Solaris 11 or later have minor version
            if ($major > 10)
            {
                # run uname -v, will have: 11.0 on solaris11, 11.1 on
                # solaris11.1
                ($rc, $output) = RunCommand('$uname -v');
                if ($rc eq '0')
                {
                    chomp $output;
                    TRACE_OUT("Output of uname -v: $output");
                    my ($major, $minor) = ($output =~ /(\d+)\.(\d+)/);
                    if (defined $minor)
                    {
                        $os_rev_minor = $minor;
                    }
                    # Cannot distinguish between OpenSolaris and Solaris
                    # starting from version 11 using 'uname -r'.
                    # But the 'uname -v' results on OpenSolaris are like
                    # snv_[version_number]b, not contain a dot, we can easily
                    # distinguish them.
                    # Note: 'uname -v' shows snv_151a on Oracle Solaris Express 11,
                    # will consider it as sol 11.0
                    # Therefore, need to check the other solaris release. 
                    else
                    {
                        DEBUG_OUT("OpenSolaris or Solaris Express detected");
                    }
                        
                }
            }
        }
        # Double check other Solaris, we do not support all OpenSolaris versions.
        # Consider Solaris Express release as the sol $os_rev_major.0 
        $output = ReadFile("/etc/release");
        chomp $output;
        TRACE_OUT("Output of /etc/release: $output");
        if (index($output, "OpenSolaris") != -1)
        {
            $os = OS->{UNKNOWN};
            $os_edition = OS_EDITION->{UNKNOWN};
            DEBUG_OUT("OpenSolaris detected, we do not support it");
        }
        elsif (index($output, "Express") != -1)
        {
            $os_rev_minor = 0;
            DEBUG_OUT("Solaris %d Express release detected.",
                $os_rev_major);
        }
    }
    return ($os, $os_rev_major, $os_rev_minor, $os_edition);
}

sub _DetectArchSolaris
{
    my $arch = ARCH->{UNKNOWN};

    my ($rc, $output) = RunCommand("$uname -p");
    if ($rc eq '0')
    {
        $arch = ARCH->{SPARC} if ($output =~ /sparc/);
        $arch = ARCH->{X86_64} if ($output =~ /i386/);  # TODO: it can also be X86 or X86_64
        # Note: Solaris also supports IA-32, PowerPC(2.5.2 only) platform
    }

    return $arch;
}


sub _DetectOSRevHpux
{
    my $os = OS->{UNKNOWN};
    my $os_rev_major = 0;
    my $os_rev_minor = 0;
    my $os_edition = OS_EDITION->{UNKNOWN};
    
    my ($rc, $output) = RunCommand("$uname -r"); # e.g., B.11.11
    if ($rc eq '0')
    {
        chomp $output;       

        my ($major, $minor) = ($output =~ m/B\.([0-9]+)\.([0-9]+)/);
        if (defined $major and defined $minor)
        {
            $os = OS->{HPUX};
            $os_rev_major = $major;
            $os_rev_minor = $minor;
            $os_edition = OS_EDITION->{STANDARD};
        }
    }
    
    return ($os, $os_rev_major, $os_rev_minor, $os_edition);
}

sub _DetectArchHpux  
{    
    my $arch = ARCH->{UNKNOWN};
    
    my ($rc, $output) = RunCommand("$uname -m");
    return ARCH->{UNKNOWN} if ($rc ne '0');

    $arch = ARCH->{PA} if ($output =~ m/9000\/7/ or $output =~ m/9000\/8/);
    $arch = ARCH->{IA64} if ($output =~ m/ia64/);

    return $arch;
}

sub _DetectOSRevDarwin
{
    my $os = OS->{UNKNOWN};
    my $os_rev_major = 0;
    my $os_rev_minor = 0;
    my $os_edition = OS_EDITION->{UNKNOWN};
    
    my ($rc, $output) = RunCommand("sw_vers -productVersion");
    
    if ($rc eq '0')
    {
        chomp $output;
        if ($output =~ m/(\d*)\.(\d*)\.?(\d*)/)
        {
            $os = OS->{DARWIN};
            $os_rev_major = $2;
            if (defined($3) && $3 != 0)
            {
                $os_rev_minor = $3;
            }
            $os_edition = OS_EDITION->{STANDARD};    
        }
    }
    
    return ($os, $os_rev_major, $os_rev_minor, $os_edition);
}

sub _DetectArchDarwin
{
    my $arch = ARCH->{UNKNOWN};    

    my ($rc, $output) = RunCommand("$uname -p");
    if ($rc eq '0')
    {
        if ($output =~ /powerpc/)
        {
            $arch = ARCH->{POWERPC};
        }
        elsif ($output =~ /i386/)
        {          
            my ($rc, $output) = RunCommand("sysctl -n hw.optional.x86_64");
            if ($rc eq '0')
            {
                chomp $output;
                if ($output == 1)
                {
                    $arch = ARCH->{X86_64};
                }
                else
                {
                    $arch = ARCH->{I386};
                }
            }
        }
    }

    return $arch;
}

#
# Get platform information
#
# Note: We consider a platform is valid only when os is NOT "unknown" and
#       os_rev_major is NOT "0".
#       We don't consider a "unknown" os_edition must lead to a "unknown" platform.
#       The platform can still have chance to be a valid platform even if 
#       the os_edition is "unknown".
#
#   return: object  - success
#           undef   - failed
#
sub GetPlatform
{   
    $uname = "uname";

    my $os_family;
    my $os;
    my $os_rev_major;
    my $os_rev_minor;
    my $os_edition;
    my $arch;
    my $platform;

    $os_family = _DetectOSFamily();
    ($os, $os_rev_major, $os_rev_minor, $os_edition) = _DetectOSRev($os_family);
    $arch = _DetectArch($os_family);

    DEBUG_OUT("OS family: [%s] OS: [%s] Version: [%s.%s] Edition: [%s] Architecture: [%s]", 
        $os_family, $os, $os_rev_major, $os_rev_minor, $os_edition, $arch);

    if (defined($os_family) and
        defined($os) and
        defined($os_rev_major) and
        defined($os_rev_minor) and
        defined($os_edition) and
        defined($arch) and
        $os_family ne OS_FAMILY->{UNKNOWN})
    {
        $platform = CentrifyDC::Platform->_new(
            $os_family, $os, $os_rev_major, $os_rev_minor, $os_edition, $arch);
    }
    else
    {
        WARN_OUT("Unknown platform detected");
        return undef;
    }

    return $platform;
}

#
# Create instance
#
#   $_[0]:  self
#   $_[1]:  OS family
#   $_[2]:  OS
#   $_[3]:  OS major revision
#   $_[4]:  OS minor revision
#   $_[5]:  OS edition
#   $_[6]:  Architecture
#
#   return: self    - success
#           undef   - failed
#
sub _new
{
    my ($class, $os_family, $os, $os_rev_major, $os_rev_minor, $os_edition, $arch) = @_;

    my $self = {
        _os_family => $os_family,
        _os => $os,
        _os_rev_major => $os_rev_major,
        _os_rev_minor => $os_rev_minor,
        _os_edition => $os_edition,
        _arch => $arch,
    };

    bless($self, $class);

    return $self;
}

#
# Get OS family
#
#   return: string (e.g. linux)
#
sub getOSFamily
{
    return $_[0]->{_os_family};
}

#
# Get OS type
#
#   return: string (e.g. rhel)
#
sub getOS
{
    return $_[0]->{_os};
}

#
# Get full OS revision
#
# This subroutine will return a string concatenation of OS type, OS edition, major version,
# and minor version (e.g. rhel_5_8_server, rhel_6_2_desktop, rhel_3_2_standard, scientific_5_10_standard, etc).
#
#   return: string  - success
#           undef   - failed
#
sub getOSRevFull
{
    my $self = $_[0];
    my $separator = '_';
    my $rc;
    if ($self->getOSEdition() eq '')
    {
        $rc = $self->getOS() . $self->getOSRevMajor() . $separator . $self->getOSRevMinor();
    }
    else
    {
        $rc = $self->getOS() . $separator . $self->getOSRevMajor() . $separator . $self->getOSRevMinor() . $separator . $self->getOSEdition();
    }
    return $rc;
}

#
# Get OS major revision
#
#   return: string
#
sub getOSRevMajor
{
    return $_[0]->{_os_rev_major};
}

#
# Get OS minor revision
#
#   return: string
#
sub getOSRevMinor
{
    return $_[0]->{_os_rev_minor};
}

#
# Get OS edition
#
#   return: string
#
sub getOSEdition
{
    return $_[0]->{_os_edition};
}

#
# Get architecture
#
#   return: string (e.g. x86_64)
#
sub getArch
{
    return $_[0]->{_arch};
}


1;
