#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2014-2020 Centrify Corporation. All rights reserved.
#
# This script is used to modify pwgr
#
# arguments:
#       please see sub usage
# return:
#       0 if succeed
#       1 if meet error
#
##############################################################################

use warnings;
use strict;
use lib "/usr/share/centrifydc/perl";
use lib "/usr/share/centrifydc/scripts";
use CentrifyDC::Include;
use Getopt::Long qw(GetOptions);

use autoedit_common_utils qw(loadSetting getValueList);
use autoedit_file_lock;
use CentrifyDC::GP::General qw(:debug :system);

sub usage
{
    warn << "END";
    This script is used to modify pwgr
    --cfg_file      the full path of pwgr
    --do_revert     set it if want to revert our changes
END
    exit 1;
}

use constant DEFAULT_CFG_FILE => "/etc/rc.config.d/pwgr";

my $cfgName = DEFAULT_CFG_FILE;
my $cfgHandler;
my $newName;
my $newHandler;

my $do_revert;
my $cdc_short_comment = "# modified by centrifydc";

GetOptions( "cfg_file=s" => \$cfgName,
            "do_revert" => \$do_revert,
            ) or usage;

my $flock = autoedit_file_lock::new("$cfgName");
if (!defined $flock)
{
    DEBUG_OUT("cannot lock file $cfgName");
    exit 1;
}

$cfgName = TraverseSymLink($cfgName);
if (!defined $cfgName)
{
    ERROR_OUT("meet error when check symlink $cfgName");
    exit 1;
}

my $cfgLines = ReadFile("$cfgName");
if (!defined $cfgLines)
{
    ERROR_OUT("cannot read $cfgName");
    exit 1;
}

if ($do_revert)
{
    $cfgLines =~ s/$cdc_short_comment\n# //g;
    $cfgLines =~ s/\n$cdc_short_comment\n.*\n//g;
}
else
{
    if ($cfgLines =~ /$cdc_short_comment/)
    {
        # it has been updated
        exit 0;
    }

    $cfgLines =~ s/^\s*PWGR\s*=\s*1\s*$/$cdc_short_comment\n# $&/mg;
    $cfgLines =~ s/^\s*PWGRD_WITH_NISPLUS\s*=\s*1\s*$/$cdc_short_comment\n# $&/mg;
}

if (!$do_revert)
{
    if ($cfgLines !~ /^\s*PWGR\s*=\s*0\s*$/m)
    {
        $cfgLines .= "\n$cdc_short_comment\nPWGR=0\n";
    }
    if ($cfgLines !~ /^\s*PWGRD_WITH_NISPLUS\s*=\s*0\s*$/m)
    {
        $cfgLines .= "\n$cdc_short_comment\nPWGRD_WITH_NISPLUS=0\n";
    }
}

my $backup = "$cfgName.pre_cdc";
if (!$do_revert)
{
    if (! CopyFile($cfgName, $backup))
    {
        ERROR_OUT("Cannot backup $cfgName");
        exit 1;
    }
}
else
{
    # install.sh and rpm/pkg scripts may override config files with
    # the backup files if found them exist.
    # To avoid affecting these code logic, we'd better delete
    # backup file after "do_revert"
    if (-e $backup)
    {
        unlink $backup;
    }
}

if (!WriteFile($cfgName, $cfgLines))
{
    ERROR_OUT("Cannot write $cfgName");
    exit 1;
}

exit 0;

