#!/bin/sh /usr/share/centrifydc/perl/run.cda

##############################################################################
#
# Copyright (C) 2014-2020 Centrify Corporation. All rights reserved.
#
# This script is used to modify login.cfg
#
# arguments:
#       please see sub usage
# return:
#       0 if succeed
#       1 if meet error (general error)
#       2 if target file is empty
#
##############################################################################

use warnings;
use strict;
use lib "/usr/share/centrifydc/perl";
use lib "/usr/share/centrifydc/scripts/da";
use CentrifyDA::Include;
use Getopt::Long qw(GetOptions);
use File::Basename;

use autoedit_file_lock;
use CentrifyDA::General qw(:debug :system);

sub usage
{
    warn << "END";
    This script is used to modify login.cfg
    --lib_name      the name of centrifydc or centrifyda lib
    --cfg_file      the full path of login.cfg
    --do_revert     set it if want to revert our changes
END
    exit 1;
}

use constant DEFAULT_CFG_FILE => "/etc/security/login.cfg";
use constant DEFAULT_CDC_TAG => "centrifydc";

my $cfgName = DEFAULT_CFG_FILE;
my $cfgHandler;
my $newName;
my $newHandler;

my $lib_name = DEFAULT_CDC_TAG;
my $cdc_shell = "/usr/bin/dzsh";
my $do_revert;

my @shell_list = ();
my @notValidShells = ("nologin", "passwd", "true", "false", "centrifyda");


#
# Guarantee there're no duplicative elements in array.
#
# $_[0]: the array ref.
#
sub unique($)
{
    my ($key) = @_;
    my %key_hash = map {$_ => 1} @{$key}; 
    @{$key} = sort keys %key_hash;
}

# get available shells from /etc/security/login.cfg
sub getAvailableShells
{
    my (@cfgLines) = @_;
    my $found_usw = 0;
    my @shell_list_temp = ();
    foreach my $line (@cfgLines)
    {
        if ($line =~ /^\s*\w+\s*:\s*$/)
        {
            $found_usw = 0;
        }
        if ($line =~ /^\s*usw\s*:/)
        {
            $found_usw = 1;
        }
        
        if ($found_usw)
        {
            if ($line =~ s/^\s*shells\s*=\s*//)
            {
                @shell_list_temp = split(/,/, $line);
            }
        }
    }

    foreach my $shell (@shell_list_temp)
    {
        # Remove things that aren't really shells
        my $flag = 0;
        foreach my $not_valid_shell (@notValidShells)
        {
            if (basename($shell) eq $not_valid_shell)
            {
                $flag = 1;
                last;
            }
        }
        
        if ($flag)
        {
            next;
        }
        
        # Remove shells of path which includes "/cdax/"
        if ($shell =~ /.*\/cdax\/.*/)
        {
            next;
        }
        
        # Exclude non-existant files
        if (! -e $shell)
        {
            next;
        }
        
        # Exclude directories
        if (-d $shell)
        {
            next;
        }
        
        push @shell_list, $shell;
    }
    # Sort and remove dups
    @shell_list = sort @shell_list;
    unique(\@shell_list);
}

GetOptions( "lib_name=s" => \$lib_name,
            "cfg_file=s" => \$cfgName,
            "do_revert" => \$do_revert,
            ) or usage;

my $flock = autoedit_file_lock::new("$cfgName");
if (!defined $flock)
{
    DEBUG_OUT("cannot lock file $cfgName");
    exit 1;
}

$cfgName = TraverseSymLink($cfgName);
if (!defined $cfgName)
{
    ERROR_OUT("meet error when check symlink $cfgName");
    exit 1;
}

my $cfgLines = ReadFile("$cfgName");
if (!defined $cfgLines)
{
    ERROR_OUT("cannot read $cfgName");
    exit 1;
}

my $cfgfile_empty = IsFileEmpty($cfgName);
if ($cfgfile_empty)
{
    exit 2;
}

my @cfgLines = split "\n", $cfgLines;

#get the shell list from /etc/security/login.cfg
getAvailableShells(@cfgLines);

# update cfg file
my @newLines = ();
my $found_usw = 0;
foreach my $line (@cfgLines)
{
    # looking for "usw:" block
    if ($line =~ /^\s*\w+\s*:\s*$/)
    {
        $found_usw = 0;
    }
    if ($line =~ /^\s*usw\s*:/)
    {
        $found_usw = 1;
    }

    if ($found_usw)
    {
        if ($line =~ /^\s*shells\s*=/)
        {
            my $cdash = "/bin/centrifyda";
            if ($do_revert)
            {
                if ($lib_name eq "centrifydc")
                {
                    $line =~ s/,\s*$cdc_shell\b//g;
                    $line =~ s/\b$cdc_shell\s*,//g;
                }
                elsif ($lib_name eq "centrifyda")
                {
                    # Now, we do not remove ../cdax/.. in the section of usw in file /etc/security/login.cfg when running "dacontrol -d"
                    # And they will be removed while uninstalling DA
                    #foreach my $shell (@shell_list)
                    #{
                    #    my $cdaxDir = dirname($shell) . "/cdax/";
                    #    $shell = $cdaxDir . basename($shell);
                    #    if ($line =~ /$shell/)
                    #    {
                    #        $line =~ s/,\s*$shell\b//g;
                    #    }
                    #}
                    
                    # Remove /bin/centrifyda when running "dacontrol -d"                   
                    if ($line =~ /$cdash/)
                    {
                        $line =~ s/,\s*$cdash\b//g;
                    }
                }
            }
            else
            {
                if ($lib_name eq "centrifydc")
                {
                    if ($line =~ /$cdc_shell/)
                    {
                        # have been using $cdc_tag
                        # nothing need to do here
                        exit 0;
                    }
                    $line .= ",$cdc_shell";
                }
                elsif ($lib_name eq "centrifyda")
                {
                    foreach my $shell (@shell_list)
                    {
                        my $cdaxDir = dirname($shell) . "/cdax/";
                        $shell = $cdaxDir . basename($shell);
                        unless ($line =~ /$shell/)
                        {
                            $line .= ",$shell";
                        }
                    }
                    
                    # Add /bin/centrifyda into shells
                    unless ($line =~ /$cdash/)
                    {
                        $line .= ",$cdash";
                    }
                }
            }
        }
    }

    push @newLines, $line;
}
my $newLines = join "\n", @newLines;
$newLines .= "\n";

my $backup = "$cfgName.pre_cdc";
if ($lib_name eq "centrifyda")
{
    $backup = "$cfgName.pre_cda";
}
if (!$do_revert)
{
    if (! CopyFile($cfgName, $backup))
    {
        ERROR_OUT("Cannot backup $cfgName");
        exit 1;
    }
}
else
{
    # install.sh and rpm/pkg scripts may override config files with
    # the backup files if found them exist.
    # To avoid affecting these code logic, we'd better delete
    # backup file after "do_revert"
    if (-e $backup)
    {
        unlink $backup;
    }
}

if (!WriteFile($cfgName, $newLines))
{
    ERROR_OUT("Cannot write $cfgName");
    exit 1;
}

exit 0;

